/* minimization of L/L#-convex function
 *  This is an implementation of steepest descent algorithm
 *  and its scaling algorithm.
 *  The calculation cost is O(2^dimension) per step.
 *  cf. K. Murota: Discrete convex analysis (Japanese), 2001, p.239-243.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"

/* returns  0 if no improvment was found. */
static long best_improve2(int dim, double func(int,int*), int x[],
                          double *min_f, const int sign, const int fix_index) {
    const unsigned long loop = 1L << dim;
    const unsigned long fix_set = (fix_index>=0) ? (1L << fix_index):(0);
    int i;
    long j, set, min_set = 0;
    double f;

    for (set=1; set<loop; set++) {
        if ((set & fix_set) != 0) set += fix_set;
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i] += sign;
        f = func(dim, x);  oracle_call++;
        if (f < *min_f) {
            *min_f = f;  min_set = set;  improve_call++;
        }
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i] -= sign;
    }
    return min_set;
}

/* returns  0 if no improvment was found. */
/* returns >0 if improvment was found in plus set */
/* returns <0 if improvment was found in minus set */
static long best_improve(int dim, double func(int,int*), int x[],
                         const int width, const int fix_index) {
    long set_plus, set_minus;
    double f, f_plus, f_minus;

    iteration++;
    f = f_plus = f_minus = func(dim, x);
    set_plus  =  best_improve2(dim, func, x, &f_plus,  +width, fix_index);
    set_minus = -best_improve2(dim, func, x, &f_minus, -width, fix_index);

    if (f_plus == f && f_minus == f) return 0;
    if (f_plus < f_minus) return set_plus;
    else                  return set_minus;
}


static void steepest_descent(int dim, double func(int,int*), int x[],
                             const int width, const int fix_index) {
    int i, sign;
    long j, set;
    int init_iteration = iteration;

    while ((set=best_improve(dim, func, x, width, fix_index)) != 0) {
        if (set > 0) sign = +1;
        else {       sign = -1;  set = -set; }
        for (i=0,j=1; j<=set; i++,j<<=1) if ((j&set) != 0) x[i]+=sign*width;
        status_print(dim, func, x);
    }
    /*
    printf("width=%d, step=%d, ", width, iteration - init_iteration);
    status_print(dim, func, x);
    */

    if (width > 1 && iteration - init_iteration > dim) {
        printf("#iteration = %d > dim\n", iteration - init_iteration);
    }
}

static double minimize(int dim, double func(int,int*), int init[],
		       const int fix_index, int width) {
    assert(LONG_MAX > pow(2, dim));
    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    while (width > 1) {
        steepest_descent(dim, func, init, width, fix_index);
        width /= 2 * sqrt(dim);  /* imploved by N.T */
    }
    steepest_descent(dim, func, init, 1, fix_index);

    counter_print();
    minimizer_print(dim, func, init);
    return func(dim, init);
}


/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
double  lconv_minimize(int dim, double func(int dim, int x[]), int init[],
		       const int fix_index) {
    return minimize(dim, func, init, fix_index, 1);
}

double lgconv_minimize(int dim, double func(int dim, int x[]), int init[]) {
    return minimize(dim, func, init, -1, 1);
}

double lconv_minimize_scaling(int dim, double func(int,int*), int init[],
                              const int fix_index) {
    int width = 20 * sqrt(dim);
    return minimize(dim, func, init, fix_index, width);
}

double lgconv_minimize_scaling(int dim, double func(int,int*), int init[]) {
    int width = 20 * sqrt(dim);
    return minimize(dim, func, init, -1, width);
}
